/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Chip type           : Atmega88 or Atmega168 or Atmega328 with ENC28J60
 *********************************************/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <avr/pgmspace.h>
#include "ip_arp_udp_tcp.h"
#include "websrv_help_functions.h"
#include "enc28j60.h"
#include "timeout.h"
#include "net.h"

// Note: This software implements a web server and a modbus client
// The web server is at "myip" and the client tries to access "msrvip".
// 
// Please modify the following lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x29};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {10,0,0,29};
// IP address of the modbus server to contact:
static uint8_t msrvip[4] = {68,62,255,136};
// Default gateway. The ip address of your DSL router. It can be set to the same as
// msrvip the case where there is no default GW to access the 
// web server (=web server is on the same lan as this host) 
static uint8_t gwip[4] = {10,0,0,2};
//
// listen port for tcp/www:
#define MYWWWPORT 80
//
// the password string, (only a-z,0-9,_ characters):
static char password[]="secret"; 
//
// -------- do not change anything below this line ---------
// global string buffer
#define STR_BUFFER_SIZE 20
static char gStrbuf[STR_BUFFER_SIZE+1];
static uint16_t gPlen;

static uint8_t modbus_dat[35];
static uint8_t have_modbus_answer=0; // len of answer
static uint8_t have_modbus_req=0; // len of req
static uint8_t modbus_client_attempts=0;
static uint8_t modbus_client_sendok=0;
// eth/ip buffer:
#define BUFFER_SIZE 590
static uint8_t buf[BUFFER_SIZE+1];

// set output to VCC, red LED off
#define LEDOFF PORTB|=(1<<PORTB1)
// set output to GND, red LED on
#define LEDON PORTB&=~(1<<PORTB1)
// to test the state of the LED
#define LEDISOFF PORTB&(1<<PORTB1)
// 

uint8_t verify_password(char *str)
{
        // the first characters of the received string are
        // a simple password/cookie:
        if (strncmp(password,str,strlen(password))==0){
                return(1);
        }
        return(0);
}

uint16_t http200ok(void)
{
        return(fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n")));
}


uint16_t print_webpage_ok(void)
{
        uint16_t plen;
        plen=http200ok();
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>OK</h2>Wait a moment before you proceed or press reload on the <a href=/>next page</a>. Modbus can be quite slow."));
        return(plen);
}

uint16_t print_webpage_modbus_req(void)
{
        uint16_t plen;
        plen=http200ok();
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>send modbus/tcp request</h2><pre>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/r method=get>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("msg  : <input type=text size=35 name=rq value=\"0,0,0,0,0,6,1,3,1,e7,0,01\">\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("passw: <input type=password size=8 name=pw>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<input type=submit value=\"send\">\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("</form>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("example:\nTID: 0,0\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("PID: 0,0\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("len: 6\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("uid: 1\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("read multi reg: 3\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("ref: 1,e7\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("wrd cnt: 0,1\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("</pre><br><hr>tuxgraphics.org"));
        return(plen);
}

// prepare the main webpage by writing the data to the tcp send buffer
uint16_t print_webpage(void)
{
        uint16_t plen;
        uint8_t i;
        plen=http200ok();
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/m>[modbus query]</a> "));
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[refresh]</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Modbus response status</h2>\n<pre>\n"));
        // convert number to string:
        plen=fill_tcp_data_p(buf,plen,PSTR("\nNumber of requests: "));
        // convert number to string:
        itoa(modbus_client_attempts,gStrbuf,10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("\nSuccessful requests: "));
        // convert number to string:
        itoa(modbus_client_sendok,gStrbuf,10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("\n\nlast response (hex):\n"));
        if (have_modbus_answer){
                i=0;
                while(i<have_modbus_answer){
                        itoa(modbus_dat[i],gStrbuf,16);
                        plen=fill_tcp_data(buf,plen,gStrbuf);
                        plen=fill_tcp_data_p(buf,plen,PSTR(" "));
                        i++;
                }
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("none"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\n</pre><br><hr>tuxgraphics.org"));
        return(plen);
}

// parse the modbus request string that the user entered in the 
// web-page. We store the data in modbus_dat and we use
// have_modbus_req and have_modbus_answer to see if modbus_dat
// contains now a request or the answer to a request.
void parse_modbus_hex(char *str){
        char *strptr2;
        strptr2=str;
        have_modbus_req=0;    // len of request
        have_modbus_answer=0; // no answer yet
        // find the end of that rq= parameter. There is either a space or a '&'
        while(*str &&  *str!=' ' && *str!='\n' && *str!='&'){
                str++;
        }
        *str='\0';// end the string
        urldecode(strptr2); // un-webify it
        // now we parse something like " 0, 8 , ab , 09,a,5 "
        str=strptr2;
        while(*str && have_modbus_req<sizeof(modbus_dat)){ // not end of string
                // find first digit:
                while(*str && ! isxdigit(*str)){
                        str++;
                }
                if (isxdigit(*str)){
                        // found a digit
                        modbus_dat[have_modbus_req]=(uint8_t)strtol(str, &strptr2, 16); 
                        have_modbus_req++;
                        // hop over the digit
                        str=strptr2;
                }
        }
}

uint16_t modbus_datafill_callback(uint8_t fd){
        fd=0; // supress warning about unused parameter
        return(fill_tcp_data_len(buf,0, modbus_dat, have_modbus_req));
}

uint8_t modbusresult_callback(uint8_t fd, uint8_t statuscode,uint16_t datapos, uint16_t len_of_data){
        have_modbus_answer=0;
        fd=0; // supress warning about unused parameter
        if (statuscode==0){
                modbus_client_sendok++;
                LEDOFF;
                // limit to size of available storage:
                if (len_of_data>sizeof(modbus_dat)){
                        len_of_data=sizeof(modbus_dat);
                }
                while(len_of_data){
                        modbus_dat[have_modbus_answer]=buf[datapos+have_modbus_answer];
                        have_modbus_answer++;
                        len_of_data--;
                }
        }
        return(1); // end the tcp dialog from the client side
}

// analyse the url given
// return values: are used to show different pages (see main program)
//
//                The string passed to this function will look like this:
//                /r?rq=0%2C0%2C0%2C0%2C0%2C6%2C1%2C3%2C3%2Ce7%2C0%2C0a&pw=secret HTTP/1.....
//                / HTTP/1.....
int8_t analyse_get_url(char *str)
{
        char *strptr;
        // the first slash:
        if (str[0] == '/' && str[1] == ' '){
                // end of url, display just the web page
                return(0);
        }
        if (strncmp("/m ",str,3)==0){
                // end of url, display just the web page
                gPlen=print_webpage_modbus_req();
                return(10);
        }
        if (strncmp("/r?",str,3)==0){
               // str is now something like /r?pw=secret&rq= 
               if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                        urldecode(gStrbuf);
                        if (verify_password(gStrbuf)){
                                strptr=strstr(str,"rq=");
                                if (strptr!=NULL){
                                        strptr+=3;
                                        parse_modbus_hex(strptr);
                                        LEDON;
                                        modbus_client_attempts++;
                                        // trigger the sending of the modbus request:
                                        client_tcp_req(&modbusresult_callback,&modbus_datafill_callback,502);
                                        return(1);
                                }
                        }
                }
        }
        return(-1); // Unauthorized
}

int main(void){
        int8_t cmd;
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(0); // 60us


        /*initialize enc28j60*/
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(0); // 60us
        
        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        
        DDRB|= (1<<DDB1); // LED, enable PB1, LED as output
        LEDOFF;

        //init the web server ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,MYWWWPORT);
        // init the web client:
        client_set_gwip(gwip);  // e.g internal IP of dsl router
        client_tcp_set_serverip(msrvip);
        //
        while(1){
                // handle ping and wait for a tcp packet
                gPlen=packetloop_icmp_tcp(buf,enc28j60PacketReceive(BUFFER_SIZE, buf));
                if(gPlen==0){
                        continue;
                }
                        
                if (strncmp("GET ",(char *)&(buf[gPlen]),4)!=0){
                        // head, post and other methods:
                        //
                        // for possible status codes see:
                        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                        gPlen=http200ok();
                        gPlen=fill_tcp_data_p(buf,gPlen,PSTR("<h1>200 OK</h1>"));
                        goto SENDTCP;
                }
                cmd=analyse_get_url((char *)&(buf[gPlen+4]));
                // for possible status codes see:
                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                if (cmd==-1){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1>"));
                        goto SENDTCP;
                }
                if (cmd==1){
                        gPlen=print_webpage_ok();
                        goto SENDTCP;
                }
                if (cmd==10){
                        goto SENDTCP;
                }
                gPlen=print_webpage();
                //
SENDTCP:
                www_server_reply(buf,gPlen); // send data

        }
        return (0);
}
